;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Module: CPUTYPEA.ASM
;
; Date: 25 July 2023
;
; Author: Michael Campbell
;
; Description:
;
; This module contains the get_cpu_type function, callable from C but
; implemented in x86 assembly language.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

.MODEL large

ct_8088         EQU 0
ct_8086         EQU 1
ct_80286        EQU 2
ct_80386        EQU 3
ct_80486        EQU 4
ct_Pentium      EQU 5
ct_PentiumPro   EQU 6
ct_Unknown      EQU 7

; Parameters that control the 8088 vs. 8086 discrimination, which involves 
; running a large number of byte and word memory accesses, and timing each.

MAX_LOOPS        EQU  4          ; Num loops = MAX_LOOPS * CYCLES_PER_LOOP
CYCLES_PER_LOOP  EQU  32000
RANGE_SIZE       EQU  1          ; Byte and Word accesses considered to be the
                                 ; the same if tick count is within +/- the range
.CODE

; .page

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUNCTION int get_cpu_type ()
;
; C calls this function via:
;   int get_cpu_type ()
;
; This function returns an integer value that represents the processor type,
; which is one of the above-defined EQU values.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

PUBLIC _GET_CPU_TYPE

_GET_CPU_TYPE PROC
.8086

LOCAL byte_target  :BYTE,  word_target  :WORD,\          ; Memory targets
      strt_ticks   :WORD,  curr_ticks   :WORD,\          ; Starting ticks
      byte_ticks   :WORD,  word_ticks   :WORD=AUTO_SIZE  ; Byte, word transfer ticks

; Set up our stack frame

   push bp   
   mov bp, sp   
   sub sp, AUTO_SIZE

   push bx                              ; Save used registers
   push cx
   push dx
   push si
   push di
   push es

;       8088/8086 CPU check. Bits 12-15 are always set on the 8088/8086
;       processor. Try to clear them. If the attempt fails, the CPU is
;       either an 8088 or an 8086.

check_8088_8086:
.8086
        pushf                           ; Get FLAGS into BX
        pop     bx
        mov     ax, 0fffh               ; Clear FLAGS bits 12-15
        and     ax, bx
        push    ax                      ; Pop revised flags back into FLAGS
        popf                            ; 
        pushf                           ; Get the resulting FLAGS into AX
        pop     ax
        and     ax, 0f000h              ; Isolate bits 12-15

        cmp     ax, 0f000h              ; If bits 12-15 are still set, the
        je      Is_8086_8088            ; CPU is either an 8088 or an 8086
        jmp     check_80286             ; Else carry on to the 286 check

Is_8086_8088:

;       OK,  we  have  either an 8088 or an 8086. The 8088 has an 8-bit
;       data  bus  while  the  8086  has  a  16-bit  data bus, so it IS
;       important  to  discriminate between the two. The only meaninful
;       difference is the speed  of  8-bit vs. 16-bit data transfers.

;       To discriminate between the two, we need to write an equal number
;       of byte and word memory transfers. On an 8086, each will take the 
;       same amount of time. On an 8088, there will be a difference between
;       the time for the 8-bit transfers and the 16-bit transfers, since 
;       each 16-bit transfers is composed of two 8-bit transfers, vs. the 
;       single 16-bit transfer that happens on an 8086.

        mov ax, 040h                          ; Load 0040h (BIOS Data Area base) into ES
        mov es, ax

        mov bx, word ptr es:[06ch]            ; Get starting BIOS ticks, lower word
        mov [strt_ticks], bx

;       Now run a large number ( MAX_LOOPS * CYCLES_PER_LOOP) of 8-bit byte
;       memory accesses

byte_loop_init:
        mov bx, MAX_LOOPS

byte_loop:
        mov cx, CYCLES_PER_LOOP

byte_cycle:
        inc byte [byte_target]
        loop short byte_cycle

        dec bx
        jnz short byte_loop

;       Byte accesses done - get the tick timer and compute ticks to do the 
;       byte accesses

        mov bx, word ptr es:[06ch]            ; Get ending BIOS ticks, lower word
        mov [curr_ticks], bx

        sub bx, [strt_ticks]                  ; Get ticks to run byte accesses
        mov [byte_ticks], bx

        jb short byte_loop_init               ; If counter has rolled over, repeat the entire test

;       Now run a large number ( MAX_LOOPS * CYCLES_PER_LOOP) of 16-bit 
;       word memory accesses

        mov bx, [curr_ticks]                  ; Set up new starting ticks
        mov [strt_ticks], bx

word_loop_init:
        mov bx, MAX_LOOPS

word_loop:
        mov cx, CYCLES_PER_LOOP

word_cycle:
        inc word [word_target]
        loop short word_cycle

        dec bx
        jnz short word_loop

        mov bx, word ptr es:[06ch]            ; Get ending BIOS ticks, lower word
        mov [curr_ticks], bx

        sub bx, [strt_ticks]                  ; Get ticks to run word accesses
        mov [word_ticks], bx

        jb short word_loop_init               ; If counter has rolled over, repeat the whole test

;       We have now timed the same number of byte and word memory accesses.
;       If this is an 8086, the tick counts for both will be the same. If
;       not, this is an 8088.

        xor dx,dx                             ; DX = 0; ticks_match = FALSE

   ;    if ( word_ticks == byte_ticks )

        mov si, [byte_ticks]
        mov di, [word_ticks]
        cmp di,si
        jne short ticks_not_equal

        mov dx,1                                  ; ticks_match = TRUE
        jmp short check_done

ticks_not_equal:
   ;
   ;    Byte and Word ticks are not equal; check if they are within the range
   ;    of +/- RANGE_SIZE.
   ;       
   ;    if ( ( word_ticks >= byte_ticks-RANGE_SIZE ) &&
   ;         ( word_ticks <= byte_ticks+RANGE_SIZE )     )
   ;           ticks_match = TRUE;
   ;    
        mov ax, si            
        sub ax, RANGE_SIZE          ; byte_ticks - range ( -2 )
        cmp di, ax
        jl  short check_done

        mov ax, si
        add ax, RANGE_SIZE          ; byte_ticks + range ( +2 )
        cmp di, ax
        jg  short check_done

        mov dx,1                    ; ticks_match = TRUE

check_done:
   ;
   ;    At this point, DX is 1 if either (a) the byte and word ticks are the
   ;    same, or (b) the byte ticks are within +/- RANGE of the word ticks.
   ;    So, if DX is 0, the CPU is an 8088. If DX is 1, it is an 8086.

        cmp dx, 0
        jnz CPU_Is_8086

CPU_Is_8088:
        mov ax, 0                             ; Load 8088 return type
        jmp end_test_cpu_type

CPU_Is_8086:
        mov ax, 1                             ; Load 8086 return type
        jmp end_test_cpu_type
       

; Intel 286 CPU check

; Bits 12-15 are always clear on the Intel 80286 processor.

check_80286:
.286
        or      bx, 0f000h              ; Try to set bits 12-15
        push    bx
        popf
        pushf
        pop     ax
        and     ax, 0f000h              ; If bits 12-15 are cleared, CPU=Intel 286
        mov     ax, 2                   ; Set the 80286 return value
        jnz     check_80386
        jmp     end_test_cpu_type


; Intel 386 CPU check

; The AC bit (bit 18) is a new bit introduced in the EFLAGS
; register on the Intel486 DX CPU to generate alignment faults.
; This bit can not be set on the Intel 386 CPU.

check_80386:
.386
        pushfd
        pop     eax                     ; Get original EFLAGS
        mov     ecx, eax                ; Save original EFLAGS
        xor     eax, 40000h             ; Flip AC bit in EFLAGS
        push    eax                     ; Save for EFLAGS
        popfd                           ; Copy to EFLAGS
        pushfd                          ; Push EFLAGS
        pop     eax                     ; Get new EFLAGS value
        xor     eax, ecx                ; Can't toggle AC bit, CPU=Intel 386
        mov     ax, 3                   ; Set the 80386 return value
        je      end_test_cpu_type

; Intel486 DX CPU, Intel 487 SX MCP, and Intel486 SX CPU checking

; Checking for the ability to set/clear the ID flag (bit 21) in EFLAGS
; which diferentiates between Pentium (or greater) and the Intel 486.
; If the ID flag is set then the CPUID instruction can be used to
; determine the final version of the chip, else it's a 486

check_80486:
.486
        pushfd                          ; Get EFLAGS into EAX
        pop     eax
        mov     ecx, eax                ; Save original EFLAGS in ECX

        xor     eax, 0200000h           ; Flip ID bit in EFLAGS
        push    eax                     ; Get modified value into EFLAGS
        popfd
        pushfd                          ; Get resulting EFLAGS into EAX
        pop     eax

        xor     eax, ecx                ; ZF=1 if they are the same (i.e. ID flag did not flip)
                                        ; So, if ZF=1, it is a late model 486 or a Pentium or later
        mov     ax, 4                   ; Set the 80486 return code

        je      end_test_cpu_type       ; if ID bit cannot be changed

;       At this point, the CPU is either a late model 486 (SL-Enhanced) or is 
;       a Pentium or later. Verify the the ID bit IS set, and if so, we can
;       use the CPUID instruction to make the final determination.

;       The below code is left in place, but there is a bug of some nature.
;       This code works perfectly every other time. On the alternate time, 
;       the initial 8088/8086 test passes! If the below code is not executed
;       this does not happen and this function works perfectly every time.
;       The VE mainline works around this by calling this function twice at
;       startup... kludgy, but gets the job done.

is_80486_Pentium:
.586
        mov     eax, 01h                ; Get CPU ID family information into EAX
        cpuid                           ; Info returned in EAX, EBX, ECX and EDX
        mov     dx, ax                  ; Save the returned CPUID signature into DX; can return it to C as part of a long vs. an int

        mov     bx, ax                  ; Get the CPU Model into BX
        and     bx, 0f0h
        shr     bx, 04h

        mov     cx, ax                  ; Get the CPU Family into CX
        and     cx, 0f00h               ; 4 = 486, 5 = Pentium, 6 = Pentium Pro, Pentium II and Pentium III
        shr     cx, 08h

        mov     ax, cx                  ; AX contains the family number

;       Sanity test the CPU family. Ensure that it is between 4 and 6 (4 = 486,
;       5 = Pentium, 6 = Pentium Pro. If below 486, set it to 486, since at 
;       this point in the algorithm, the CPU is a 486 or above. If above 
;       Pentium Pro, set it to Pentium Pro, since that is the highest that
;       the VE C source code supports.

check_lt_4:                             ; Check if less than 4; if so make it 4
        cmp     ax, 4
        jae     check_gt_5
        mov     ax, 4                   ; Force result = 486

check_gt_5:                             ; Check if greater than 6; if so make it 6
        cmp     ax, 5
        jbe     end_test_cpu_type
        mov     ax, 5                   ; Default to Pentium if > supported
                                        ; by VE C code
end_test_cpu_type:

.8086
   ; clean up our stack frame

   pop es
   pop di
   pop si
   pop dx
   pop cx
   pop bx

   mov sp, bp
   pop bp

   ; AX is loaded with the CPU Type indicator, which is where C looks for 
   ; integer returned values, so off we go

   ret

_GET_CPU_TYPE   ENDP

END

